#
# ademco.core - Security system base classes
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyn

# constants and parameters
MAX_PARTITIONS = 8

# zone state bits
ZS_CLOSED	= 0
ZS_OPEN		= 1
ZS_TROUBLE	= 2
ZS_ALARM	= 4
ZS_BYPASSED = 8
ZS_UNKNOWN	= 0x10		# internal

# arming modes
ARM_AWAY	= 'AA'
ARM_STAY	= 'AH'
ARM_INSTANT	= 'AI'
ARM_MAX		= 'AM'
ARM_FORCE_AWAY = 'FA'
ARM_FORCE_STAY = 'FH'
ARM_UNREADY	= '??'
ARM_BYPASSED = '?B'
DISARM		= 'XX'


#
# A panel "user" (by number, with pin)
#
class User(object):

	def __init__(self, uid, pin):
		self.uid = uid
		self.pin = pin


#
# Canonical command framing
#
def frame(cmd):
	""" Frame an outgoing message according to the Ademco protocol. """
	return ("%02.2X" % (len(cmd) + 6)) + cmd + '00'

def unframe(cmd):
	""" Check and remove framing from an incoming message.

		Returns (code, data) or None if invalid.
	"""
	# LLxxxxx00CK -> xxxxx
	try:
		c_len = int(cmd[0:2], 16)
		pad = cmd[-4:-2]
		c_cksum = cmd[-2:]
		if c_len == len(cmd) and pad == '00' and cmd[-2:] == cksum(cmd[0:-2]):
			return (cmd[2:4], cmd[4:-4])	# trim length, "00", and cksum
	except ValueError:
		pass

def cksum(msg):
	""" Calculate the canonical Ademco message checksum. """
	cksum = 0
	for c in msg:
		cksum += ord(c)
	return "%02.2X" % ((-cksum) & 0xFF)


#
# Various objects that belong to a security system
#
class SystemPart(asyn.Callable):

	def __init__(self, system, callout=None):
		asyn.Callable.__init__(self, callout=callout)
		self.system = system


#
# A virtual keypad.
# All keypads assigned to a given partition act as mirrors.
# Thus, operating on a virtual keypad is the same as working
# a physical one for the same partition.
#
KEYPAD_LEN = 32

class Keypad(SystemPart):
	""" A simulated standard security panel keypad display. Read-only. """

	LED_READY	= 1
	LED_TROUBLE	= 2
	LED_ARMED	= 4

	def __init__(self, partition, callout=None):
		SystemPart.__init__(self, partition.system, callout=callout)
		self.partition = partition
		self.text = " " * KEYPAD_LEN
		self.led = 0
		self.backlight = False
		self.raw = self.text + "0"

	def update(self, body):
		old_raw = self.raw
		self.raw = body
		self.backlight = ord(body[0]) & 0x80
		self.text = chr(ord(body[0]) & ~0x80) + body[1:KEYPAD_LEN]
		self.led = int(body[KEYPAD_LEN:], 16)
		if body != old_raw:
			self.callout('change', self)

	def type(self, text):
		maxper=5
		for i in xrange(0, len(text), maxper):
			self.system.command('KS%s%s' % (self.partition.pn, text[i:i+maxper]))

	def read(self):
		def complete(ctx, *args):
			if ctx.state == 'keypad':
				self.update(*args)
		self.system.command('KD%d' % self.partition.pn, callout=complete)

	@property
	def ready_led(self): return bool(self.led & self.LED_READY)
	@property
	def trouble_led(self): return bool(self.led & self.LED_TROUBLE)
	@property
	def armed_led(self): return bool(self.led & self.LED_ARMED)

	def __repr__(self):
		return "<pad P%d %s>" % (self.partition.pn, self.raw)

	def __str__(self):
		s = "P%d" % self.partition.pn
		if self.backlight:
			s += "[BL]"
		if self.ready_led:
			s += "(ready)"
		if self.trouble_led:
			s += "(TROUBLE)"
		if self.armed_led:
			s += "(armed)"
		return s + " " + self.text


#
# Partitions are optional, and are "named" by strings '1' through '8'.
#
class Partition(SystemPart):

	def __init__(self, system, pn):
		SystemPart.__init__(self, system)
		self.pn = int(pn)
		assert self.pn > 0 and self.pn <= MAX_PARTITIONS
		self.name = 'Partition %d' % self.pn	# default
		self.status = None
		self.kp_display = ""
		self.kp_led = 0

	AST = {
		'A': ARM_AWAY,
		'H': ARM_STAY,
		'D': DISARM,
		'N': ARM_UNREADY,
		'M': ARM_MAX,
		'I': ARM_INSTANT,
		'-': ARM_UNREADY,		# undocumented
	}

	def updateStatus(self, sc):
		old_status = self.status
		self.status = self.AST.get(sc) or sc
		if old_status and self.status != old_status:
			self.callout('change', self, self.status, old_status)

	def __str__(self):
		return self.name

	def __repr__(self):
		return "<part%d %s>" % (self.pn, self.name)


#
# A sensor zone on a system.
# This includes current (dynamic) state.
#
class Zone(SystemPart):
	""" A sensor zone on a given security system.

	"""

	def __init__(self, system, zn, part):
		SystemPart.__init__(self, system)
		self.control = system.control
		self.zn = int(zn)
		self.partition = part
		self.name = "zone %d" % self.zn
		self.description = self.name
		self.status = ZS_UNKNOWN

	def updateStatus(self, state):
		old_status = self.status
		self.status = state
		if old_status != ZS_UNKNOWN and old_status != self.status:
			self.callout('change', self, self.status, old_status)

	@property
	def text_status(self):
		return self.status_to_text(self.status)

	@staticmethod
	def status_to_text(st):
		states = []
		if st & ZS_OPEN:
			states.append("open")
		if st & ZS_TROUBLE:
			states.append("TROUBLE")
		if st & ZS_ALARM:
			states.append("ALARM")
		if st & ZS_BYPASSED:
			states.append("bypassed")
		if st & ZS_UNKNOWN:
			states.append("unknown")
		return '/'.join(states) if states else "closed"

	def __repr__(self):
		return "<zone%d[%s] %s %r>" % (self.zn, self.status, self.description, self.partition)

	def __str__(self):
		return "%s(%s)" % (self.description, self.text_status)
