#
# Honeywell/Ademco alarm panel interface.
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyn
import cyin.devstate
import ademco
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *


ARMING_TYPE = dict(
	away = ademco.ARM_AWAY,
	stay = ademco.ARM_STAY,
	instant = ademco.ARM_INSTANT,
	max = ademco.ARM_MAX
)


#
# A security system device
#
class System(cyin.devstate.IPDevice, cyin.devstate.Idler):
	""" An Indigo device representing an entire security system. """

	speed = cyin.PluginProperty(type=int, default=1200)

	@property
	def system(self): return self.target

	DEFAULT_PORT = 4999
	def SERIAL_CONFIG(self):
		return dict(baudrate=self.speed)

	def connected(self, socket):
		self.target = ademco.System(cyin.plugin, callout=self._event)
		self.proceed("exploring")
		self.system.open(socket)
		self.idle_update()

	@action
	def arm_many(self, action):
		type = ARMING_TYPE.get(action.type)
		if type:
			self.system.arm(type, action.auth(), action.partition_devices)
		else:
			error(self.name, 'invalid arming mode "%s"' % type)

	@action
	def disarm_many(self, action):
		self.system.disarm(action.auth(), action.partition_devices)

	@action
	def raw_command(self, action):
		try:
			self.system.command(action.command)
		except Exception as e:
			error(f'{self.name} cannot send command "{action.command}": {e}')

	def _event(self, ctx, *args):
		if ctx.error:
			return self.fail_soft(ctx)
		if ctx.state == 'discover':
			type = args[0]
			if type == 'partition':
				debug('discovered partition', args[1].pn)
			elif type == 'done':
				debug('discovery complete')
				self.proceed("ready", recovered=True)
		elif ctx.state == 'ack':
			pass
		elif ctx.state == 'change':
			pass
		elif ctx.state == 'event':
			pass	# raw event tap
		elif ctx.state == 'received':
			(code, body) = args
			for event in Received.all_matching(self, code):
				if event.lastinput:
					event.lastinput_var = "%s%s" % (code, body)
				event.trigger_me()
		elif ctx.state == 'END':
			self.reconnect()
		elif ctx.state == 'close':
			pass
		elif ctx.state == 'timeout':
			self.fail_soft("no response from device")
		else:
			debug("unrecognized event", ctx, *args)


#
# Common features of all parts of a security system
#
class SystemPart(cyin.devstate.SubDevice):

	address = cyin.PluginProperty(eval=True)


#
# A system partition
#
class Partition(SystemPart):

	PARTNAME = "partition"

	def set_subaddress(self, subaddr):
		self.pno = int(subaddr)

	def setup(self, ctx=None):
		system = self.hostdev.system
		if self.pno not in system.partition:
			self.fail_hard("no partition %d in %s" % (self.pno, self.hostdev.name))
			return
		self.partition = system.partition[self.pno]
		self.partition.add_callout(self._event)
		self.update(self.partition.status, recovered=True)

	def update(self, new, recovered=False):
		if new in [ademco.ARM_AWAY, ademco.ARM_MAX]:
			self.proceed("away", recovered=recovered)
		elif new in [ademco.ARM_STAY, ademco.ARM_INSTANT]:
			self.proceed("stay", recovered=recovered)
		elif new == ademco.ARM_UNREADY:
			self.proceed("unready", recovered=recovered)
		elif new == ademco.DISARM:
			self.proceed("ready", recovered=recovered)
		else:
			error("unrecognized partition state", new)
			self.proceed("unavailable")		# but don't fail

	@action
	def arm(self, action):
		type = ARMING_TYPE.get(action.type)
		if type:
			self.partition.system.arm(type, action.auth(), [self.partition])
		else:
			error(self.name, 'invalid arming mode "%s"' % type)

	@action
	def disarm(self, action):
		self.partition.system.disarm(action.auth(), [self.partition])

	def user(self, action):
		return ademco.User(action.user, action.code)

	def _event(self, ctx, *args):
		if ctx.state == 'change':
			part, new, old = args
			self.update(new)
		else:
			debug('unrecognized', self.name, 'event', ctx, *args)

	def __repr__(self):
		return "<partition %d of %s>" % (self.pno, self.hostdev)


#
# A sensor zone
#
class Zone(SystemPart):

	onOffState = cyin.DeviceState(type=bool)

	PARTNAME = "zone"

	def set_subaddress(self, subaddr):
		self.zno = int(subaddr)

	def setup(self, ctx=None):
		self.system = self.hostdev.system
		if self.zno not in self.system.zone:
			self.fail_hard("no zone %d in %s" % (self.zno, self.hostdev.name))
			return
		self.zone = self.system.zone[self.zno]
		self.zone.add_callout(self._event)
		self.update(self.zone.status, recovered=True)

	zsmap = {
		ademco.ZS_CLOSED: "ready",
		ademco.ZS_OPEN: "open",
		ademco.ZS_TROUBLE: "trouble",
		ademco.ZS_ALARM: "alarm",
		ademco.ZS_BYPASSED: "bypassed",
	}

	@action
	def standard_status(self):
		""" Relay standard status requests """
		self.system.zone_status()

	def update(self, new, recovered=False):
		self.proceed(self.zsmap.get(new, "trouble"), recovered=recovered)
		self.onOffState = new not in [ademco.ZS_CLOSED, ademco.ZS_BYPASSED]
		if new == ademco.ZS_ALARM:
			error(self.name, "ALARM")
		elif new == ademco.ZS_TROUBLE:
			error(self.name, "is in trouble")

	def _event(self, ctx, *args):
		if ctx.state == 'change':
			zno, new, old = args
			self.update(new)
		else:
			debug('unrecognized', self.name, 'event', ctx, *args)

	def __repr__(self):
		return "<zone %d of %s>" % (self.zno, self.hostdev)


#
# A Monitor is a programmed watcher of zones
#
class Monitor(cyin.devstate.Device):

	active_list = cyin.DeviceState(type=str)

	device = cyin.PluginProperty(type=cyin.device)
	zones = cyin.PluginProperty(type=lambda v: map(cyin.device, v), untype=indigo.List)

	isopen = cyin.PluginProperty(type=bool, reconfigure='notify')
	istrouble = cyin.PluginProperty(type=bool, reconfigure='notify')
	isbypassed = cyin.PluginProperty(type=bool, reconfigure='notify')
	isalarm = cyin.PluginProperty(type=bool, reconfigure='notify')

	def start(self):
		super(Monitor, self).start()
		self.set_hostdev(self.device)
		self.set_display_address(self.hostdev.name)
		if self.hostdev and self.hostdev.ready():
			self.setup()

	def setup(self, ctx=None):
		system = self.hostdev.system
		system.add_callout(self._event)
		self.update(recovered=True)

	def stop(self):
		if self.hostdev and self.hostdev.system:
			self.hostdev.system.remove_callout(self._event, required=False)
		super(Monitor, self).stop()

	def config_changed(self, changed):
		""" Update state if one of our is* properties has changed. """
		self.update()
		super(Monitor, self).config_changed(changed)

	def _event(self, ctx, *args):
		if ctx.state == 'event':
			self.update()

	def contains(self, zone):
		""" Check whether a given zone is in our alert set """
		state = zone.state
		if state == "open":		return self.isopen
		if state == "trouble":	return self.istrouble
		if state == "bypassed":	return self.isbypassed
		if state == "alarm":	return self.isalarm
		# all other states never match

	def update(self, recovered=False):
		""" Update our state """
		zones = [z for z in self.zones if self.contains(z)]
		zones.sort(key=lambda z: z.name)		# sort by zone name
		if zones:
			self.proceed("alert", recovered=recovered)
			self.active_list = ', '.join([zone.name for zone in zones])
		else:
			self.proceed("ready", recovered=recovered)
			self.active_list = ""


#
# A virtual keypad device
#
class Keypad(SystemPart):

	display = cyin.DeviceState(type=str)
	ready_led = cyin.DeviceState(type=bool)
	armed_led = cyin.DeviceState(type=bool)
	trouble_led = cyin.DeviceState(type=bool)

	PARTNAME = "partition"

	def set_subaddress(self, subaddr):
		self.pno = int(subaddr)

	def setup(self, ctx=None):
		system = self.hostdev.system
		if self.pno not in system.partition:
			self.fail_hard("no partition %d in %s" % (self.pno, self.hostdev.name))
			return
		self.partition = system.partition[self.pno]
		self.keypad = ademco.Keypad(self.partition, callout=self._event)
		self.keypad.read()
		self.proceed("ready", recovered=True)

	def _event(self, ctx, *args):
		if ctx.error:
			self.fail_hard(ctx.error)
		if ctx.state == 'change':
			self._update()

	def _update(self):
		self.display = self.keypad.text
		self.ready_led = self.keypad.ready_led
		self.trouble_led = self.keypad.trouble_led
		self.armed_led = self.keypad.armed_led

	@action
	def keypad_type(self, action):
		self.keypad.type(action.text)
		self.keypad.read()

	@action
	def keypad_update(self, action):
		self.keypad.read()


#
# Actions
#
class _Arming(cyin.DeviceAction):
	user = cyin.PluginProperty(type=int, eval=True)
	code = cyin.PluginProperty(eval=True,
		check=[check_format(r'\d{4}', error='!Arming codes must be exactly four digits.')])

	def auth(self):
		return ademco.User(self.user, self.code)

class Arm(_Arming):
	type = cyin.PluginProperty()

class Disarm(_Arming):
	pass

class _Arming_Many(_Arming):
	partlist = cyin.PluginProperty(name="partitions", type=list)
	
	@property
	def partitions(self):
		return [cyin.device(devid) for devid in self.partlist]

	@property
	def partition_devices(self):
		return [dev.partition for dev in self.partitions]
	

class Arm_Many(_Arming_Many):
	type = cyin.PluginProperty()

class Disarm_Many(_Arming_Many):
	pass


class Keypad_Type(cyin.DeviceAction):
	text = cyin.PluginProperty(type=str, eval=True)

class Keypad_Update(cyin.DeviceAction):
	pass


class Raw_Command(cyin.DeviceAction):
	command = cyin.PluginProperty(type=str, eval=True)
	class UI(cyin.ConfigUI):
		def description(self):
			return 'send %s to %s' % (self.command, self.device.name)


#
# Events
#
class Received(cyin.Event):
	""" A raw message-received event, filtered by device and major code """

	device = cyin.PluginProperty(type=cyin.device, required=False)
	code = cyin.PluginProperty(type=str, required=False, check=[check_format('.{2}', error="Response codes must be exactly two characters long.")])
	lastinput = cyin.PluginProperty(required=False, reconfigure=False)
	lastinput_var = cyin.NamedVariable(name="lastinput")

	def matches(self, system, code):
		if self.device:
			if self.device != system:
				return False
		if self.code:
			if self.code != code:
				return False
		return True


#
# Menu filters
#
class PartitionFilter(cyin.MenuGenerator):
	""" Menu filter - ademco partitions in known System objects. """
	def generate(self):
		for system in System.all():
			for part in system.system.partition.values():
				yield Partition.subfilter_clause(system, part.pn)

class ZoneFilter(cyin.MenuGenerator):
	""" Menu filter - ademco zones in known System objects. """
	def generate(self):
		for system in System.all():
			for zone in system.system.zone.values():
				yield Zone.subfilter_clause(system, zone.zn)

class ZoneDeviceFilter(cyin.MenuGenerator):
	""" Menu filter - Indigo Zone devices in a given System. """
	def generate(self):
		dev = self.ui.device
		for zone in Zone.all():
			if zone.ready() and zone.hostdev == dev:
				yield zone.filter_clause()

class PartitionDeviceFilter(cyin.MenuGenerator):
	""" Menu filter - Indigo Partition devices in a given System. """
	def generate(self):
		dev = self.ui.device
		for part in Partition.all():
			if part.ready() and part.hostdev == dev:
				yield part.filter_clause()


#
# The plugin.
#
class Plugin(cyin.asynplugin.Plugin):
	pass
